"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras.src.datasets.imdb import get_word_index
from keras.src.datasets.imdb import load_data
