"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""


from keras.src.initializers import deserialize
from keras.src.initializers import get
from keras.src.initializers import serialize
from keras.src.initializers.constant_initializers import Constant
from keras.src.initializers.constant_initializers import Constant as constant
from keras.src.initializers.constant_initializers import Identity
from keras.src.initializers.constant_initializers import Identity as IdentityInitializer
from keras.src.initializers.constant_initializers import Identity as identity
from keras.src.initializers.constant_initializers import Ones
from keras.src.initializers.constant_initializers import Ones as ones
from keras.src.initializers.constant_initializers import Zeros
from keras.src.initializers.constant_initializers import Zeros as zeros
from keras.src.initializers.initializer import Initializer
from keras.src.initializers.random_initializers import GlorotNormal
from keras.src.initializers.random_initializers import GlorotNormal as glorot_normal
from keras.src.initializers.random_initializers import GlorotUniform
from keras.src.initializers.random_initializers import GlorotUniform as glorot_uniform
from keras.src.initializers.random_initializers import HeNormal
from keras.src.initializers.random_initializers import HeNormal as he_normal
from keras.src.initializers.random_initializers import HeUniform
from keras.src.initializers.random_initializers import HeUniform as he_uniform
from keras.src.initializers.random_initializers import LecunNormal
from keras.src.initializers.random_initializers import LecunNormal as lecun_normal
from keras.src.initializers.random_initializers import LecunUniform
from keras.src.initializers.random_initializers import LecunUniform as lecun_uniform
from keras.src.initializers.random_initializers import OrthogonalInitializer
from keras.src.initializers.random_initializers import OrthogonalInitializer as Orthogonal
from keras.src.initializers.random_initializers import OrthogonalInitializer as orthogonal
from keras.src.initializers.random_initializers import RandomNormal
from keras.src.initializers.random_initializers import RandomNormal as random_normal
from keras.src.initializers.random_initializers import RandomUniform
from keras.src.initializers.random_initializers import RandomUniform as random_uniform
from keras.src.initializers.random_initializers import TruncatedNormal
from keras.src.initializers.random_initializers import TruncatedNormal as truncated_normal
from keras.src.initializers.random_initializers import VarianceScaling
from keras.src.initializers.random_initializers import VarianceScaling as variance_scaling
